"""Public request common classes and utils."""

from datetime import datetime
from enum import Enum, unique
from typing import Optional

from fred_types import BaseId, BaseModel
from pydantic import Field


class PublicRequestId(BaseId):
    """Public request unique identifier."""


@unique
class RegistryObjectType(str, Enum):
    """Object type enumeration."""

    CONTACT = "contact"
    DOMAIN = "domain"
    KEYSET = "keyset"
    NSSET = "nsset"


class Reference(BaseModel):
    """Object reference - type and unique identifier."""

    type: RegistryObjectType
    value: str = Field(..., min_length=1)


@unique
class RequestStatus(str, Enum):
    """Public request processing status."""

    OPEN = "opened"
    RESOLVED = "resolved"
    INVALIDATED = "invalidated"


@unique
class RequestOnStatusAction(str, Enum):
    """Public request action status."""

    SCHEDULED = "scheduled"
    PROCESSED = "processed"
    FAILED = "failed"


class PublicRequest(BaseModel):
    """Public request info.

    Attributes:
        public_request_id: unique identifier of public request
        numeric_id: unique identifier of public request (db index)
        reference: type (domain, ...) and unique identifier of referenced object
        type: public request type (same as PublicRequestType.name)
        registrar_id: sponsoring registrar holding linked object
        response_email: contact for receiving request resolution answer
        create_time: time and log item of request creation
        creation_id: log item of request creation
        resolve_time: time of request resolution
        resolving_id: log item of request resolution
        status: request resolution status
        on_status_action: request processing status
    """

    public_request_id: PublicRequestId
    numeric_id: int
    reference: Reference
    type: str
    registrar_id: Optional[str] = None
    response_email: Optional[str] = None
    create_time: datetime
    creation_id: Optional[str] = None
    resolve_time: Optional[datetime] = None
    resolving_id: Optional[str] = None
    status: RequestStatus
    on_status_action: RequestOnStatusAction
