from unittest import IsolatedAsyncioTestCase
from unittest.mock import sentinel

from fred_api.public_request.common_types_pb2 import (
    CannotBeProcessed as CannotBeProcessedMsg,
    PublicRequestDoesNotExist as PublicRequestDoesNotExistMsg,
)
from fred_api.public_request.service_public_request_processing_grpc_pb2 import (
    InvalidateReply as InvalidateReplyMsg,
    InvalidateRequest as InvalidateRequestMsg,
    ResolveReply as ResolveReplyMsg,
    ResolveRequest as ResolveRequestMsg,
)
from frgal.utils import AsyncTestClientMixin, make_awaitable

from prequel.common import PublicRequestId
from prequel.exceptions import CannotBeProcessed, PublicRequestDoesNotExist
from prequel.processing import PublicRequestProcessingClient


class TestPublicRequestProcessingClient(AsyncTestClientMixin, PublicRequestProcessingClient):
    """Test client for PublicRequestProcessingClient."""


class PublicRequestProcessingTest(IsolatedAsyncioTestCase):
    def setUp(self):
        self.client = TestPublicRequestProcessingClient(netloc=sentinel.netloc)

    async def test_resolve_ok(self):
        public_request_id = PublicRequestId("ID-1")
        resolving_id = "Resolved OK"
        self.client.mock.return_value = make_awaitable(ResolveReplyMsg())
        await self.client.resolve(public_request_id, resolving_id)
        request = ResolveRequestMsg()
        request.public_request_id.value = public_request_id
        request.resolving_id.value = resolving_id
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.PublicRequest.Api.PublicRequestProcessing/resolve",
            timeout=None,
        )

    async def test_resolve_not_found(self):
        reply = ResolveReplyMsg()
        reply.exception.public_request_does_not_exist.CopyFrom(PublicRequestDoesNotExistMsg())
        self.client.mock.return_value = make_awaitable(reply)
        with self.assertRaises(PublicRequestDoesNotExist):
            await self.client.resolve(PublicRequestId("ID-1"), "Resolved OK")

    async def test_resolve_cannot(self):
        reply = ResolveReplyMsg()
        reply.exception.cannot_be_processed.CopyFrom(CannotBeProcessedMsg())
        self.client.mock.return_value = make_awaitable(reply)
        with self.assertRaises(CannotBeProcessed):
            await self.client.resolve(PublicRequestId("ID-1"), "Resolved OK")

    async def test_invalidate_ok(self):
        public_request_id = PublicRequestId("ID-1")
        resolving_id = "Invalidated"
        self.client.mock.return_value = make_awaitable(InvalidateReplyMsg())
        await self.client.invalidate(public_request_id, resolving_id)
        request = InvalidateRequestMsg()
        request.public_request_id.value = public_request_id
        request.resolving_id.value = resolving_id
        self.client.mock.assert_called_once_with(
            request,
            method="/Fred.PublicRequest.Api.PublicRequestProcessing/invalidate",
            timeout=None,
        )

    async def test_invalidate_not_found(self):
        reply = InvalidateReplyMsg()
        reply.exception.public_request_does_not_exist.CopyFrom(PublicRequestDoesNotExistMsg())
        self.client.mock.return_value = make_awaitable(reply)
        with self.assertRaises(PublicRequestDoesNotExist):
            await self.client.invalidate(PublicRequestId("ID-1"), "Invalidated")

    async def test_invalidate_cannot(self):
        reply = InvalidateReplyMsg()
        reply.exception.cannot_be_processed.CopyFrom(CannotBeProcessedMsg())
        self.client.mock.return_value = make_awaitable(reply)
        with self.assertRaises(CannotBeProcessed):
            await self.client.invalidate(PublicRequestId("ID-1"), "Invalidated")
